package de.hft.stuttgart.citydoctor2.healing;

import static org.junit.Assert.assertFalse;
import static org.junit.Assert.assertNotNull;
import static org.junit.Assert.assertTrue;

import java.io.File;
import java.io.IOException;
import java.util.ArrayList;
import java.util.List;

import org.citygml4j.xml.writer.CityGMLWriteException;
import org.junit.Rule;
import org.junit.Test;
import org.junit.rules.TemporaryFolder;

import de.hft.stuttgart.citydoctor2.check.CheckError;
import de.hft.stuttgart.citydoctor2.check.Checker;
import de.hft.stuttgart.citydoctor2.check.Requirement;
import de.hft.stuttgart.citydoctor2.check.RequirementConfiguration;
import de.hft.stuttgart.citydoctor2.check.ValidationConfiguration;
import de.hft.stuttgart.citydoctor2.datastructure.BoundarySurface;
import de.hft.stuttgart.citydoctor2.datastructure.Building;
import de.hft.stuttgart.citydoctor2.datastructure.CityDoctorModel;
import de.hft.stuttgart.citydoctor2.datastructure.Geometry;
import de.hft.stuttgart.citydoctor2.exceptions.CityDoctorWriteException;
import de.hft.stuttgart.citydoctor2.healer.Healer;
import de.hft.stuttgart.citydoctor2.parser.CityGmlParseException;
import de.hft.stuttgart.citydoctor2.parser.CityGmlParser;
import de.hft.stuttgart.citydoctor2.parser.InvalidGmlFileException;

public class TestHealer {

	@Rule
	public TemporaryFolder folder = new TemporaryFolder();
	
	@Test
	public void testMissingMainHealing() throws CityGmlParseException, IOException, InvalidGmlFileException, CityDoctorWriteException {
		String path = "src/test/resources/LoD2_33_396_5807_AmWaldrand_fix_with_addresses.gml";
		ValidationConfiguration config = ValidationConfiguration.loadStandardValidationConfig();
		config.setSchematronFilePathInGlobalParameters("src/test/resources/missingMainBuilding.xml");
		CityDoctorModel model = TestUtil.loadCityModel(path, config);

		Checker c = new Checker(config, model);
		c.runChecks();
		
		Building b = null;
		for (Building building : model.getBuildings()) {
			if (building.getGmlId().getGmlString().equals("DEBE08YYF0004KZs")) {
				b = building;
				break;
			}
		}
		assertNotNull(b);
		
		List<CheckError> errors = new ArrayList<>();
		b.collectContainedErrors(errors);
		assertFalse(errors.isEmpty());
		
		Healer healer = new Healer(c);
		
		healer.healCityObject(b);
		assertFalse(b.containsAnyError());
	}

	@Test
	public void testHealingAPI()
			throws CityGmlParseException, IOException, InvalidGmlFileException, CityDoctorWriteException {
		File in = new File("src/test/resources/TestBuilding1.gml");
		File config = new File("src/test/resources/testConfigWithStreaming.yml");
		File out = folder.newFile("out.gml");
		out.delete();
		Healer.heal(in, config, out);
		assertTrue(out.exists());
	}

	@Test
	public void testHealingMainWithStreaming() throws IOException, CityGmlParseException, InterruptedException,
			InvalidGmlFileException, CityDoctorWriteException {
		File in = new File("src/test/resources/TestBuilding1.gml");
		File configFile = new File("src/test/resources/streamingStandardConfig.yml");
		File out = folder.newFile("out.gml");
		out.delete();

		String[] args = new String[6];
		args[0] = "-in";
		args[1] = in.getAbsolutePath();
		args[2] = "-out";
		args[3] = out.getAbsolutePath();
		args[4] = "-config";
		args[5] = configFile.getAbsolutePath();
		Healer.main(args);

		assertTrue(out.exists());

		ValidationConfiguration valConfig = ValidationConfiguration.loadStandardValidationConfig();
		valConfig.setSchematronFilePathInGlobalParameters(null);
		CityDoctorModel model = CityGmlParser.parseCityGmlFile(out.getAbsolutePath(), valConfig.getParserConfiguration());
		Checker c = new Checker(valConfig, model);
		c.runChecks();
		model.createFeatureStream().forEach(co -> {
			List<CheckError> errors = new ArrayList<>();
			co.collectContainedErrors(errors);
			assertFalse(co.containsAnyError());
		});
	}

	/**
	 * Building has a lot of consecutive points same errors
	 * 
	 * @throws CityGmlParseException
	 * @throws IOException
	 * @throws CityGMLBuilderException
	 * @throws CityGMLWriteException
	 * @throws InvalidGmlFileException
	 */
	@Test
	public void testHealing1()
			throws CityGmlParseException, IOException, CityGMLWriteException, InvalidGmlFileException {
		String path = "src/test/resources/TestBuilding1.gml";
		ValidationConfiguration config = ValidationConfiguration.loadStandardValidationConfig();
		CityDoctorModel model = TestUtil.loadCityModel(path, config);

		Checker c = new Checker(config, model);
		c.runChecks();

		Healer healer = new Healer(c);
		Geometry geom = model.getBuildings().get(0).getGeometries().get(0);
		healer.healCityObject(model.getBuildings().get(0));

		c.runChecks();
		assertFalse(geom.containsAnyError());
	}

	/**
	 * Many consecutive points same and manifold edge errors
	 * 
	 * @throws CityGmlParseException
	 * @throws IOException
	 * @throws CityGMLBuilderException
	 * @throws CityGMLWriteException
	 * @throws InvalidGmlFileException
	 */
	@Test
	public void testHealing2()
			throws CityGmlParseException, IOException, CityGMLWriteException, InvalidGmlFileException {
		String path = "src/test/resources/TestBuilding2.gml";
		ValidationConfiguration config = ValidationConfiguration.loadStandardValidationConfig();
		CityDoctorModel model = TestUtil.loadCityModel(path, config);

		Checker c = new Checker(config, model);
		c.runChecks();

		Geometry geom = model.getBuildings().get(0).getGeometries().get(0);

		Healer healer = new Healer(c);
		healer.setNumberOfIterations(33);
		healer.healCityObject(model.getBuildings().get(0));
		c.runChecks();

		assertFalse(geom.containsAnyError());
	}

	@Test
	public void testHealing3()
			throws CityGmlParseException, IOException, CityGMLWriteException, InvalidGmlFileException {
		String path = "src/test/resources/TestBuilding3.gml";
		ValidationConfiguration config = ValidationConfiguration.loadStandardValidationConfig();
		CityDoctorModel model = TestUtil.loadCityModel(path, config);

		Checker c = new Checker(config, model);
		c.runChecks();

		Geometry geom = model.getBuildings().get(0).getGeometries().get(0);

		Healer healer = new Healer(c);
		healer.healCityObject(model.getBuildings().get(0));
		c.runChecks();

		assertFalse(geom.containsAnyError());
	}

	@Test
	public void testHealing4()
			throws CityGmlParseException, IOException, CityGMLWriteException, InvalidGmlFileException {
		String path = "src/test/resources/TestBuilding4.gml";
		ValidationConfiguration config = ValidationConfiguration.loadStandardValidationConfig();
		CityDoctorModel model = TestUtil.loadCityModel(path, config);

		Checker c = new Checker(config, model);
		c.runChecks();

		Geometry geom = model.getBuildings().get(0).getGeometries().get(0);

		Healer healer = new Healer(c);
		healer.setNumberOfIterations(83);
		healer.healCityObject(model.getBuildings().get(0));
		c.runChecks();

		assertFalse(geom.containsAnyError());
	}

	@Test
	public void testHealing5()
			throws CityGmlParseException, IOException, CityGMLWriteException, InvalidGmlFileException {
		String path = "src/test/resources/TestBuilding5.gml";
		ValidationConfiguration config = ValidationConfiguration.loadStandardValidationConfig();
		CityDoctorModel model = TestUtil.loadCityModel(path, config);

		Checker c = new Checker(config, model);
		c.runChecks();

		Geometry geom = model.getBuildings().get(0).getGeometries().get(0);

		Healer healer = new Healer(c);
		healer.setNumberOfIterations(165);
		healer.healCityObject(model.getBuildings().get(0));
		c.runChecks();

		assertFalse(geom.containsAnyError());
	}

	@Test
	public void testHealing6()
			throws CityGmlParseException, IOException, CityGMLWriteException, InvalidGmlFileException {
		String path = "src/test/resources/TestBuilding6.gml";
		ValidationConfiguration config = ValidationConfiguration.loadStandardValidationConfig();
		RequirementConfiguration checkConfiguration = config.getRequirements()
				.get(Requirement.R_SE_BS_IS_CEILING.toString());
		checkConfiguration.setEnabled(false);
		checkConfiguration = config.getRequirements().get(Requirement.R_SE_BS_IS_FLOOR.toString());
		checkConfiguration.setEnabled(false);
		checkConfiguration = config.getRequirements().get(Requirement.R_SE_BS_IS_GROUND.toString());
		checkConfiguration.setEnabled(false);
		checkConfiguration = config.getRequirements().get(Requirement.R_SE_BS_IS_WALL.toString());
		checkConfiguration.setEnabled(false);
		checkConfiguration = config.getRequirements().get(Requirement.R_SE_BS_ROOF_UNFRAGMENTED.toString());
		checkConfiguration.setEnabled(false);
		CityDoctorModel model = TestUtil.loadCityModel(path, config);

		Checker c = new Checker(config, model);
		c.runChecks();

		Building b = model.getBuildings().get(0);
		Healer healer = new Healer(c);
		healer.healCityObject(b);
		c.runChecks();

		for (BoundarySurface bs : b.getBoundarySurfaces()) {
			assertFalse(bs.containsAnyError());
		}
	}

	@Test
	public void testHealing8()
			throws CityGmlParseException, IOException, CityGMLWriteException, InvalidGmlFileException {
		String path = "src/test/resources/TestBuilding8.gml";
		ValidationConfiguration config = ValidationConfiguration.loadStandardValidationConfig();
		CityDoctorModel model = TestUtil.loadCityModel(path, config);

		Checker c = new Checker(config, model);
		c.runChecks();

		Geometry geom = model.getBuildings().get(0).getGeometries().get(0);

		Healer healer = new Healer(c);
		healer.setNumberOfIterations(45);
		healer.healCityObject(model.getBuildings().get(0));
		c.runChecks();

		assertFalse(geom.containsAnyError());
	}
	
	@Test
	public void testHealingFailWrite() throws CityGmlParseException, IOException, InvalidGmlFileException, CityDoctorWriteException {
		String path = "src/test/resources/TestBuilding8.gml";
		ValidationConfiguration config = ValidationConfiguration.loadStandardValidationConfig();
		CityDoctorModel model = TestUtil.loadCityModel(path, config);

		Checker c = new Checker(config, model);
		c.runChecks();


		Healer healer = new Healer(c);
		healer.setNumberOfIterations(45);
		healer.healModel(model);
	}
	
	@Test
	public void testHealingBigBuildingWithoutErrors()
			throws CityGmlParseException, IOException, CityGMLWriteException, InvalidGmlFileException {
		String path = "src/test/resources/bigBuilding1.gml";
		ValidationConfiguration config = ValidationConfiguration.loadStandardValidationConfig();
		config.setSchematronFilePathInGlobalParameters(null);
		CityDoctorModel model = TestUtil.loadCityModel(path, config);

		Checker c = new Checker(config, model);
		c.runChecks();
		
		Geometry geom = model.getBuildings().get(0).getGeometries().get(0);
		Healer healer = new Healer(c);
		healer.setNumberOfIterations(100);
		healer.healCityObject(model.getBuildings().get(0));
		
		c.runChecks();
		assertFalse(geom.containsAnyError());
	}
}
